//
//  Renderer.swift
//  Solar System
//
//  Created by Albert Kwaśkiewicz on 5/4/20.
//  Copyright © 2020 Albert Kwaśkiewicz. All rights reserved.
//

import Foundation
import MetalKit

class Renderer: NSObject{
    
    static var device: MTLDevice!
    static var commandQueue: MTLCommandQueue!
    static var library: MTLLibrary!
    var mesh: MTKMesh!
    
    var pipelineState: MTLRenderPipelineState!
    let depthStencilState: MTLDepthStencilState
    
    var uniforms = Uniforms()
    var fragmentUniforms = FragmentUniforms()
    
    lazy var camera: Camera = {
        let camera = ArcballCamera()
        camera.distance = 3
        camera.target = [0, 0, 0]
        camera.rotation.x = Float(-10).degreesToRadians
        return camera
    }()
    
//    var models: [Model] = []
    var planets: [Planet] = []
    var sun: Sun
    //    var meshes: [MTKMesh] = []
    
    // MARK: INIT
    init(metalView: MTKView) {
        guard let device = MTLCreateSystemDefaultDevice(),
            let commandQueue = device.makeCommandQueue() else { fatalError("GPU not available") }
        
        Renderer.device = device
        Renderer.commandQueue = commandQueue
        Renderer.library = device.makeDefaultLibrary()
        metalView.device = device
        metalView.depthStencilPixelFormat = .depth32Float
        
        pipelineState = Renderer.buildPipelineState()
        depthStencilState = Renderer.buildDepthStencilState()!
        
        sun = Sun()
        
        super.init()
        metalView.clearColor = MTLClearColor(red: 1.0, green: 0.5, blue: 0.5, alpha: 1)
        metalView.delegate = self
        
//        let mercury = Model(name: "mercury.obj")
//        mercury.position = [3.0, 0, 0]
//        mercury.rotation = [0, 0, 0]
//        models.append(mercury)
        
        //        _ = sun.createSun(device: device)
        let earth = Planet(planet: .earth)
        let mars = Planet(planet: .mars)
        //        planets.append(sun)
        planets.append(earth)
//        planets.append(mars)
        
        for planet in planets {
            planet.position = [sun.radius / sun.scaleFactor + planet.distance / planet.scaleFactor, 0, 0]
            print("Planet: \(planet.name) of size \(planet.size) with meshes count of \(planet.meshes.count)")
        }
        
        mtkView(metalView, drawableSizeWillChange: metalView.bounds.size)
    }
    
    static func buildPipelineState() -> MTLRenderPipelineState {
        let library = Renderer.library
        let vertexFunction = library?.makeFunction(name: "vertex_main")
        let fragmentFunction = library?.makeFunction(name: "fragment_main")
        
        var pipelineState: MTLRenderPipelineState
        let pipelineDescriptor = MTLRenderPipelineDescriptor()
        pipelineDescriptor.vertexFunction = vertexFunction
        pipelineDescriptor.fragmentFunction = fragmentFunction
        let vertexDescriptor = MDLVertexDescriptor.defaultVertexDescriptor
        pipelineDescriptor.vertexDescriptor = MTKMetalVertexDescriptorFromModelIO(vertexDescriptor)
        pipelineDescriptor.colorAttachments[0].pixelFormat = .bgra8Unorm
        pipelineDescriptor.depthAttachmentPixelFormat = .depth32Float
        do {
            pipelineState = try Renderer.device.makeRenderPipelineState(descriptor: pipelineDescriptor)
        } catch let error {
            fatalError(error.localizedDescription)
        }
        return pipelineState
    }
    
    static func buildDepthStencilState() -> MTLDepthStencilState? {
        let descriptor = MTLDepthStencilDescriptor()
        descriptor.depthCompareFunction = .less
        descriptor.isDepthWriteEnabled = true
        
        return Renderer.device.makeDepthStencilState(descriptor: descriptor)
    }
}

extension Renderer: MTKViewDelegate {
    func mtkView(_ view: MTKView, drawableSizeWillChange size: CGSize) {
        camera.aspect = Float(view.bounds.width) / Float(view.bounds.height)
    }
    
    // MARK: DRAW
    func draw(in view: MTKView) {
        guard let descriptor = view.currentRenderPassDescriptor,
            let commandBuffer = Renderer.commandQueue.makeCommandBuffer(),
            let renderEncoder = commandBuffer.makeRenderCommandEncoder(descriptor: descriptor)
            else { return }
        
        renderEncoder.setDepthStencilState(depthStencilState)
        
        uniforms.projectionMatrix = camera.projectionMatrix
        uniforms.viewMatrix = camera.viewMatrix
        fragmentUniforms.cameraPosition = camera.position
      print(uniforms.projectionMatrix)
        for planet in planets {
            
            uniforms.modelMatrix = planet.modelMatrix
          uniforms.modelMatrix = float4x4.identity()
            uniforms.normalMatrix = planet.modelMatrix.upperLeft
            
            renderEncoder.setVertexBytes(&uniforms,
                                         length: MemoryLayout<Uniforms>.stride,
                                         index: Int(BufferIndexUniforms.rawValue))
            renderEncoder.setRenderPipelineState(planet.pipelineState)
            
            for mesh in planet.meshes {
              print("rendering ", mesh.name)
                let vertexBuffer = mesh.vertexBuffers[0].buffer
                renderEncoder.setVertexBuffer(vertexBuffer,
                                              offset: 0,
                                              index: Int(BufferIndexVertices.rawValue))
                
                for submesh in mesh.submeshes {
                    renderEncoder.drawIndexedPrimitives(type: .triangle,
                                                        indexCount: submesh.indexCount,
                                                        indexType: submesh.indexType,
                                                        indexBuffer: submesh.indexBuffer.buffer,
                                                        indexBufferOffset: submesh.indexBuffer.offset)
                }
            }
        }
        
//        for model in models {
//            uniforms.modelMatrix = model.modelMatrix
//            uniforms.normalMatrix = model.modelMatrix.upperLeft
//
//            renderEncoder.setVertexBytes(&uniforms,
//                                         length: MemoryLayout<Uniforms>.stride,
//                                         index: Int(BufferIndexUniforms.rawValue))
//            renderEncoder.setRenderPipelineState(model.pipelineState)
//
//            for mesh in model.meshes {
//                let vertexBuffer = mesh.mtkMesh.vertexBuffers[0].buffer
//                renderEncoder.setVertexBuffer(vertexBuffer, offset: 0, index: Int(BufferIndexVertices.rawValue))
//
//                for submesh in mesh.submeshes {
//                    let mtkMesh = submesh.mtkSubmesh
//
//                    renderEncoder.drawIndexedPrimitives(type: .triangle,
//                                                        indexCount: mtkMesh.indexCount,
//                                                        indexType: mtkMesh.indexType,
//                                                        indexBuffer: mtkMesh.indexBuffer.buffer,
//                                                        indexBufferOffset: mtkMesh.indexBuffer.offset)
//                }
//            }
//        }
        
        renderEncoder.endEncoding()
        guard let drawable = view.currentDrawable else { return }
        commandBuffer.present(drawable)
        commandBuffer.commit()
    }
}
